#include "ipobject.h"
#include <QTcpServer>
#include <QTcpSocket>
#ifndef QT_NO_SYSTEMSEMAPHORE
#include <QSystemSemaphore>
#endif
#include <QCoreApplication>
#include <QTime>
#include "globals.h"

IPObject::IPObject(const QString &name, const quint16 port, QObject *parent) :
		QObject(parent), port(port), mutex(QMutex::Recursive), bLinkInProgress(false)
{
	client = new QTcpSocket(this);
	server = new QTcpServer(this);
#ifndef QT_NO_SYSTEMSEMAPHORE
	lock = new QSystemSemaphore(name, 1, QSystemSemaphore::Create);
#endif

	connect(client, SIGNAL(disconnected()), this, SLOT(link()));
	connect(client, SIGNAL(error(QAbstractSocket::SocketError)), this, SLOT(link()));
	connect(client, SIGNAL(readyRead()), this, SLOT(receive()));
	connect(server, SIGNAL(newConnection()), this, SLOT(processConnection()));

	link();
}

IPObject::~IPObject()
{
#ifndef QT_NO_SYSTEMSEMAPHORE
	delete lock;
#endif
}

void IPObject::handleDisconnection()
{
	emit clientLeft();
}

void IPObject::processConnection()
{
	if (!server->isListening() || !server->hasPendingConnections())
		return;
	qDebug() << "incomming connection";
	QTcpSocket *sock = NULL;
	while ((sock = server->nextPendingConnection()))
	{
		connect(sock, SIGNAL(disconnected()), this, SLOT(handleDisconnection()));
		connect(sock, SIGNAL(disconnected()), sock, SLOT(deleteLater()));
		connect(sock, SIGNAL(readyRead()), this, SLOT(broadcast()));
	}
}

void IPObject::link()
{
	if (server->isListening())
		return;

	if (client->state() != QTcpSocket::UnconnectedState)
		return;

	if (bLinkInProgress)
		return;

	mutex.lock();
	bLinkInProgress = true;
	mutex.unlock();

#ifndef QT_NO_SYSTEMSEMAPHORE
	lock->acquire();
#endif

	client->connectToHost(QHostAddress::LocalHost, port);
	if (!client->waitForConnected(2000))
	{
		server->listen(QHostAddress::LocalHost, port);
	}

#ifndef QT_NO_SYSTEMSEMAPHORE
	lock->release();
#endif

	mutex.lock();
	bLinkInProgress = false;
	mutex.unlock();

	if (isServing())
	{
		emit enterServerMode();
#ifndef QT_NO_DEBUG
		qDebug() << "entered server mode";
#endif
	}
	else
	{
		emit enterClientMode();
#ifndef QT_NO_DEBUG
		qDebug() << "entered client mode";
#endif
	}
}

bool IPObject::isServing() const
{
	return server->isListening();
}

void IPObject::broadcast()
{
	if (!isServing())
		return;
	foreach(QObject *obj, server->children())
	{
		QTcpSocket *sock = dynamic_cast<QTcpSocket*>(obj);
		if (!sock)
			continue;
		if (sock->state() != QTcpSocket::ConnectedState)
			continue;
		while(sock->bytesAvailable())
		{
			int len = 0;
			if (sock->read((char*)&len, sizeof(int)) != sizeof(int))
				break;
			QByteArray data = sock->read(len);
			len -= data.length();
			QTime timer;
			timer.start();
			while(len && timer.elapsed() < 10000)
			{
				sock->waitForReadyRead();
				QByteArray tmp = sock->read(len);
				len -= tmp.length();
				data.append(tmp);
			}
			send(data);
			receive(data);
		}
	}
}

void IPObject::receive()
{
	while(client->bytesAvailable())
	{
		int len = 0;
		if (client->read((char*)&len, sizeof(len)) != sizeof(len))
			break;
		QByteArray data = client->read(len);
		len -= data.length();
		QTime timer;
		timer.start();
		while(len > 0 && timer.elapsed() < 10000)
		{
			client->waitForReadyRead();
			QByteArray tmp = client->read(len);
			len -= tmp.length();
			data.append(tmp);
		}
		receive(data);
	}
}

void IPObject::send(const QByteArray &message)
{
	const int len = message.size();
	if (!isServing())
	{
		if (client->state() == QTcpSocket::ConnectedState)
		{
			while (client->bytesToWrite() > 0)
				client->waitForBytesWritten();
			client->write((const char*)&len, sizeof(int));
			qint64 l = client->write(message);
			QTime timer;
			timer.start();
			while(l < len && timer.elapsed() < 10000)
			{
				while (client->bytesToWrite() > 0)
					client->waitForBytesWritten();
				l += client->write(message.right(len - l));
			}
		}
		return;
	}
	foreach(QObject *obj, server->children())
	{
		QTcpSocket *sock = dynamic_cast<QTcpSocket*>(obj);
		if (!sock)
			continue;
		if (sock->state() != QTcpSocket::ConnectedState)
			continue;
		while(sock->bytesToWrite() > 0)
			sock->waitForBytesWritten();
		sock->write((const char*)&len, sizeof(int));
		qint64 l = sock->write(message);
		QTime timer;
		timer.start();
		while(l < len && timer.elapsed() < 10000)
		{
			while (sock->bytesToWrite() > 0)
				sock->waitForBytesWritten();
			l += sock->write(message.right(len - l));
		}
	}
}

void IPObject::send(const QString &message)
{
	QByteArray data;
	QDataStream stream(&data, QIODevice::WriteOnly);
	stream << message;
	send(data);
}

void IPObject::send(const char *message)
{
	send(QString(message));
}

void IPObject::waitForData(int msec)
{
	if (!isServing())
	{
		if (client->bytesAvailable() == 0)
			client->waitForReadyRead(msec);
		if (client->bytesAvailable() > 0)
			receive();
	}
	else
	{
		bool ready = false;
		foreach(QObject *obj, server->children())
		{
			QTcpSocket *sock = dynamic_cast<QTcpSocket*>(obj);
			if (!sock)
				continue;
			if (sock->state() != QTcpSocket::ConnectedState)
				continue;
			if (sock->bytesAvailable() == 0)
				sock->waitForReadyRead(msec);
			ready |= sock->bytesAvailable() > 0;
		}
		if (ready)
			broadcast();
	}
}
