#include "mainwindow.h"
#include "ui_mainwindow.h"
#include "about.h"
#include "config.h"
#include "trayicon.h"
#include "stats.h"
#include "searchfile.h"
#include "ipcsocket.h"
#include <QSettings>
#include "globals.h"
#include "sharedpath.h"
#include "ipfileserver.h"
#include <QDebug>

using namespace qshare;

MainWindow::MainWindow(QWidget *parent)
    : QMainWindow(parent), ui(new Ui::MainWindow)
{
    ui->setupUi(this);

	NetDiscovery::instance()->setParent(this);
	new Config(this);
	new TrayIcon(this);
	SharedPath::instance()->setParent(this);
	IPFileServer::instance()->setParent(this);

	ipcsock = new IPCSocket(this);
	connect(ipcsock,SIGNAL(connected()),this,SLOT(checkIPCStatus()));
	connect(ipcsock,SIGNAL(message(QByteArray)),this,SLOT(replyToIPCMessage(QByteArray)));
#ifdef Q_OS_UNIX
	ipcsock->connectTo("qShare" + qgetenv("USER"));
#else
	ipcsock->connectTo("qShare");
#endif

	connect(ui->action_AboutQt, SIGNAL(triggered()), this, SLOT(aboutQt()));
	connect(ui->action_About, SIGNAL(triggered()), this, SLOT(about()));
	connect(ui->action_Config, SIGNAL(triggered()), Config::instance(), SLOT(show()));
	connect(ui->action_Exit, SIGNAL(triggered()), this, SLOT(close()));
	connect(ui->action_Exit, SIGNAL(triggered()), qApp, SLOT(quit()));
	connect(NetDiscovery::instance(), SIGNAL(serverListChanged()), this, SLOT(refreshServerList()));
	connect(ui->bRefresh, SIGNAL(clicked()), NetDiscovery::instance(), SLOT(forceRefresh()));
	connect(ui->bOpen, SIGNAL(clicked()), this, SLOT(openServer()));
	connect(ui->serverList, SIGNAL(itemDoubleClicked(QListWidgetItem*)), this, SLOT(openServer()));
	connect(ui->actionStatistics, SIGNAL(triggered()), this, SLOT(showStatsDialog()));
	connect(ui->action_Find,SIGNAL(triggered()), this, SLOT(showSearchFileDialog()));

	refreshServerList();
    setWindowTitle(windowTitle() + ' ' + tr("on") + ' ' + Config::instance()->getHostname());
}

MainWindow::~MainWindow()
{
    delete ui;
}

void MainWindow::showSearchFileDialog()
{
	SearchFile::instance()->show();
}

void MainWindow::showStatsDialog()
{
	Stats::instance()->show();
}

void MainWindow::about()
{
	About::instance()->show();
}

void MainWindow::aboutQt()
{
	QApplication::aboutQt();
}

void MainWindow::refreshServerList()
{
	lServer = NetDiscovery::instance()->getServerList();
	QList<QListWidgetItem*> selection = ui->serverList->selectedItems();
	QString selectedName = selection.isEmpty() ? QString() : selection.first()->text();

	ui->serverList->clear();
	foreach(Computer server, lServer)
	{
		QListWidgetItem *item;
		ui->serverList->addItem(item = new QListWidgetItem(QIcon(":/icons/computer"), server.getName()));
		if (server.getName() == selectedName)
			item->setSelected(true);
	}
}

void MainWindow::changeEvent(QEvent *e)
{
	QMainWindow::changeEvent(e);
	switch (e->type()) {
	case QEvent::LanguageChange:
		ui->retranslateUi(this);
        setWindowTitle(windowTitle() + ' ' + tr("on") + ' ' + Config::instance()->getHostname());
		break;
	default:
		break;
	}
}

void MainWindow::closeEvent(QCloseEvent *e)
{
	settings->beginGroup("mainwindow");
	settings->setValue("size", size());
	settings->setValue("pos", pos());
	settings->endGroup();

	QMainWindow::closeEvent(e);
}

void MainWindow::showEvent(QShowEvent *e)
{
	QMainWindow::showEvent(e);

	settings->beginGroup("mainwindow");
	resize(settings->value("size", size()).toSize());
	move(settings->value("pos", pos()).toPoint());
	settings->endGroup();
}

void MainWindow::openServer()
{
	QList<QListWidgetItem*> items = ui->serverList->selectedItems();

	foreach(QListWidgetItem* item, items)
	{
		foreach(Computer c, lServer)
		{
			if (c.getName() == item->text())
				c.open();
		}
	}
}

void MainWindow::checkIPCStatus()
{
	if (ipcsock->state() == IPCSocket::Client)		// qShare is already started
	{
		// Parse command line arguments and send them to the running instance
		QStringList args = QCoreApplication::arguments();
		// Show main window by default (if no arguments is given)
		if (args.size() == 1)
			ipcsock->sendMessage("MainWindow::show()");
		args.pop_front();		// Remove program name
		foreach(QString arg, args)
			ipcsock->sendMessage(arg.toUtf8());
		qApp->processEvents();
		exit();
	}

	if (ipcsock->state() == IPCSocket::Server)
	{
		// Parse command line arguments and send them to replyToIPCMessage
		QStringList args = QCoreApplication::arguments();
		args.pop_front();		// Remove program name
		foreach(QString arg, args)
			replyToIPCMessage(arg.toUtf8());
	}
}

void MainWindow::replyToIPCMessage(const QByteArray &msg)
{
	if (msg == "MainWindow::show()")
		show();
	else if (msg == "About::show()")
		About::instance()->show();
	else if (msg == "Config::show()")
		Config::instance()->show();
	else if (msg == "Stats::show()")
		Stats::instance()->show();
	else if (msg == "SearchFile::show()")
		SearchFile::instance()->show();
	else if (msg == "MainWindow::hide()")
		hide();
	else if (msg == "About::hide()")
		About::instance()->hide();
	else if (msg == "Config::hide()")
		Config::instance()->hide();
	else if (msg == "Stats::hide()")
		Stats::instance()->hide();
	else if (msg == "SearchFile::hide()")
		SearchFile::instance()->hide();
	else if (msg == "MainWindow::exit()" || msg == "exit")
		exit();
	else if (msg.startsWith("--share="))
	{
		const QByteArray _path = msg.right(msg.size() - 8);
		const QString path = QString::fromUtf8(_path.data(), _path.size());
		Config::instance()->addPath(false, path);		// By default don't add writable paths
	}
	else if (msg.startsWith("--unshare="))
	{
		const QByteArray _path = msg.right(msg.size() - 10);
		const QString path = QString::fromUtf8(_path.data(), _path.size());
		Config::instance()->removePath(path);
	}
}

void MainWindow::exit()
{
	QTimer *timer = new QTimer;
	timer->setInterval(1);
	timer->setSingleShot(false);
	connect(timer,SIGNAL(timeout()),qApp,SLOT(quit()));
	timer->start();
}
