#include "netdiscovery.h"
#include <QUdpSocket>
#include <QStringList>
#include <QThreadPool>
#include "config.h"
#include "delay.h"
#include "bonjourregistrar.h"
#include <QApplication>

NetDiscovery *NetDiscovery::pInstance = NULL;

NetDiscovery *NetDiscovery::instance()
{
	if (pInstance)
		return pInstance;
	return pInstance = new NetDiscovery;
}

NetDiscovery::NetDiscovery() : IPObject("qShareNetDiscovery", 8956), mutex(QMutex::Recursive), bRunning(false), registrar(NULL)
{
	connect(this, SIGNAL(enterServerMode()), SLOT(start()));

	setAutoDelete(false);
	bStop = false;
	bRefreshForced = false;
	bSendNotifyRequest = false;

	start();
	forceRefresh();
}

NetDiscovery::~NetDiscovery()
{
	terminate();
	while(isRunning())
		Delay::msleep(1);
	if (registrar)
		delete registrar;
}

void NetDiscovery::run()
{
	if (!isServing())
		return;

	mutex.lock();
	if (bRunning)
	{
		mutex.unlock();
		return;
	}

	bRunning = true;
	mutex.unlock();

	mServer.clear();
	emit serverListChanged();

	QUdpSocket sock;

	sock.bind(2000);

	bSendNotifyRequest = true;

#ifdef USE_ZEROCONF
	if (registrar)
		delete registrar;
	registrar = new BonjourRegistrar(this);
	registrar->registerService(BonjourRecord(qApp->applicationName() + " " + qApp->applicationVersion() + " on " + Config::getHostname(),
											 "_ftp._tcp", ""),
							   Config::getServerPort());
#endif

	quint32 n = -1;
	int nTO = 0;
	while(!hasToStop() && isServing())
	{
		Delay::msleep(1000);
		++n;
		while(sock.hasPendingDatagrams())
		{
			char buf[1024];
			QHostAddress host;
			memset(buf, 0, sizeof(buf));
			sock.readDatagram(buf, 1024, &host);
			setOrUpdate(buf, host);
		}

		if (n >= 10 || bRefreshForced)
		{
			bRefreshForced = false;
			QString msg = Config::getHostname() + '\n' + QString::number(Config::getServerPort());
			sock.writeDatagram(msg.toAscii(), QHostAddress(QHostAddress::Broadcast), 2000);
			n = 0;
		}

		if (bSendNotifyRequest)
		{
			bSendNotifyRequest = false;
			sock.writeDatagram("REFRESH", QHostAddress(QHostAddress::Broadcast), 2000);
		}

		++nTO;
		if (nTO >= 2)
		{
			QVector<QString> vDead;
			for(QMap<QString, Computer>::iterator i = mServer.begin() ; i != mServer.end() ; ++i)
				if (i->timedOut())
					vDead << i.key();
			if (!vDead.isEmpty())
			{
				for(QVector<QString>::const_iterator i = vDead.begin() ; i != vDead.end() ; ++i)
					mServer.remove(*i);
				emit serverListChanged();
			}
			nTO = 0;
		}
	}

#ifdef USE_ZEROCONF
	if (registrar)
		delete registrar;
	registrar = NULL;
#endif

	mutex.lock();
	bStop = false;
	bRunning = false;
	mutex.unlock();
}

void NetDiscovery::terminate()
{
	mutex.lock();
	bStop = true;
	mutex.unlock();
}

void NetDiscovery::setOrUpdate(const QString& data, const QHostAddress &host)
{
	if (!isServing())
		return;

	QStringList params = data.split('\n', QString::SkipEmptyParts);
	if (params.size() < 2)
	{
		if (params.size() == 1 && params.front() == "REFRESH")
			bRefreshForced = true;
		return;
	}

	Computer server;
	server.setHostname(host.toString());
	server.setName(params[0]);
	server.setPort(params[1].toInt());

	bool changed = false;
	mutex.lock();
	QString key = host.toString();
	changed = !mServer.contains(key) || mServer[key] != server;
	mServer[key] = server;
	send("SET", key, server);
	mutex.unlock();
	if (changed)
		emit serverListChanged();
}

void NetDiscovery::remove(const QHostAddress &host)
{
	if (isServing())
		send("REMOVE", host.toString());
	mutex.lock();
	if (mServer.contains(host.toString()))
	{
		mServer.remove(host.toString());
		mutex.unlock();
		if (isServing())
			emit serverListChanged();
	}
	else
		mutex.unlock();
}

QList<Computer> NetDiscovery::getServerList()
{
	QList<Computer> lServer;

	mutex.lock();
	for(QMap<QString, Computer>::iterator i = mServer.begin() ; i != mServer.end() ; ++i)
		lServer.push_back(i.value());
	mutex.unlock();

	return lServer;
}

void NetDiscovery::forceRefresh()
{
	if (!isServing())
	{
		mServer.clear();
		emit serverListChanged();
		send("forceRefresh");
		return;
	}
	mutex.lock();
	mServer.clear();
	bRefreshForced = true;
	bSendNotifyRequest = true;
	emit serverListChanged();
	mutex.unlock();
}

Computer NetDiscovery::getComputer(const QString &name)
{
	QMutexLocker mLock(&mutex);
	for(QMap<QString, Computer>::iterator i = mServer.begin() ; i != mServer.end() ; ++i)
		if (i.value().getName() == name)
			return i.value();
	return Computer();
}

void NetDiscovery::receive(const QByteArray &message)
{
	QDataStream stream(message);
	QString type;
	stream >> type;
	if (type == "SIG serverListChanged")
	{
		if (isServing())
			return;
		emit serverListChanged();
		return;
	}
	if (type == "forceRefresh")
	{
		if (isServing())
			forceRefresh();
		return;
	}
	if (type == "REMOVE")
	{
		QString host;
		stream >> host;
		remove(QHostAddress(host));
		return;
	}
	if (type == "SET")
	{
		QString name;
		Computer computer;
		stream >> name >> computer;
		bool bChange = !mServer.contains(name) || mServer[name] != computer;
		mServer[name] = computer;
		if (bChange)
			emit serverListChanged();
		return;
	}
}

void NetDiscovery::propagateServerListChanged()
{
	send("SIG serverListChanged");
}

void NetDiscovery::start()
{
	setAutoDelete(false);
	QThreadPool::globalInstance()->start(this, QThread::LowestPriority);
}

bool NetDiscovery::hasToStop()
{
	QMutexLocker mLock(&mutex);
	return bStop;
}

bool NetDiscovery::isRunning()
{
	QMutexLocker mLock(&mutex);
	return bRunning;
}
