#include "searchfile.h"
#include "ui_searchfile.h"
#include "netdiscovery.h"
#include "searchthread.h"
#include "sharedfolders.h"
#include "config.h"

SearchFile *SearchFile::pInstance = NULL;

SearchFile *SearchFile::instance()
{
	if (pInstance)
		return pInstance;
	return new SearchFile();
}

void SearchFile::deleteInstance()
{
	if (pInstance)
		pInstance->deleteLater();
}

SearchFile::SearchFile(QWidget *parent) :
    QDialog(parent),
    m_ui(new Ui::SearchFile)
{
	pInstance = this;

    m_ui->setupUi(this);

	m_ui->bSearch->setVisible(true);
	m_ui->bSearch->setEnabled(true);
	m_ui->bStop->setVisible(false);
	m_ui->bStop->setEnabled(false);
	m_ui->expression->setEnabled(true);

	connect(m_ui->bSearch,SIGNAL(clicked()),this,SLOT(startSearch()));
	connect(m_ui->bStop,SIGNAL(clicked()),this,SLOT(stopSearch()));
	connect(m_ui->results,SIGNAL(cellDoubleClicked(int,int)),this,SLOT(openResult(int,int)));
	timer = new QTimer(this);
	timer->setInterval(500);
	timer->setSingleShot(true);
	connect(timer,SIGNAL(timeout()),this,SLOT(updateResults()));
	workingThreads = 0;
    setWindowTitle(windowTitle() + " (" + tr("from") + ' ' + Config::instance()->getHostname() + ')');
}

SearchFile::~SearchFile()
{
    delete m_ui;

	pInstance = NULL;
}

void SearchFile::hideEvent(QHideEvent *)
{
	deleteLater();
}

void SearchFile::changeEvent(QEvent *e)
{
    QDialog::changeEvent(e);
    switch (e->type()) {
    case QEvent::LanguageChange:
        m_ui->retranslateUi(this);
        setWindowTitle(windowTitle() + " (" + tr("from") + ' ' + Config::instance()->getHostname() + ')');
        break;
    default:
        break;
    }
}

void SearchFile::startSearch()
{
	if (workingThreads > 0)
		return;
	m_ui->bSearch->setEnabled(false);
	m_ui->bSearch->setVisible(false);
	m_ui->bStop->setEnabled(true);
	m_ui->bStop->setVisible(true);
	m_ui->expression->setEnabled(false);

	// Clear the results
	lock();
	m_ui->results->setRowCount(0);
	unlock();

	QString request = m_ui->expression->text();
	QList<Computer> serverList = NetDiscovery::instance()->getServerList();
	workingThreads = 0;
	for(int i = 0 ; i < serverList.size() ; ++i)
	{
		SearchThread *thread = new SearchThread(serverList[i].getName(), request, this);
		connect(thread,SIGNAL(destroyed()),this,SLOT(threadFinished()));
		connect(this,SIGNAL(stoppingSearch()),thread,SLOT(stopSearch()));
		connect(thread,SIGNAL(found(QString,QString)),this,SLOT(addResult(QString,QString)));
		++workingThreads;
		thread->start();
	}
}

void SearchFile::stopSearch()
{
	emit stoppingSearch();

	m_ui->bSearch->setVisible(true);
	m_ui->bSearch->setEnabled(true);
	m_ui->bStop->setVisible(false);
	m_ui->bStop->setEnabled(false);
	m_ui->expression->setEnabled(true);
}

void SearchFile::addResult(const QString &computer, const QString &path)
{
	QMutexLocker mLock(this);

	queue.enqueue(QPair<QString, QString>(computer, path));
	if (!timer->isActive())
		timer->start();

}

void SearchFile::threadFinished()
{
	QMutexLocker mLock(this);
	--workingThreads;
	if (workingThreads == 0)
		stopSearch();
}

void SearchFile::updateResults()
{
	QMutexLocker mLock(this);

	int cur = m_ui->results->rowCount();
	m_ui->results->setRowCount(cur + queue.size());
	while(!queue.isEmpty())
	{
		QPair<QString,QString> file = queue.dequeue();
		QTableWidgetItem *item;
		m_ui->results->setItem(cur,0,item = new QTableWidgetItem(QIcon(":/icons/computer.png"), file.first));
		item->setFlags(Qt::ItemIsSelectable | Qt::ItemIsEnabled);
		m_ui->results->setItem(cur,1,item = new QTableWidgetItem(file.second));
		item->setFlags(Qt::ItemIsSelectable | Qt::ItemIsEnabled);
		++cur;
	}
	m_ui->results->sortItems(1,Qt::AscendingOrder);
}

void SearchFile::openResult(int row, int)
{
	lock();
	QString computerName = m_ui->results->item(row, 0)->text();
	QString filename = m_ui->results->item(row, 1)->text();
	unlock();

	Computer computer = NetDiscovery::instance()->getComputer(computerName);
	computer.open(SharedFolders::getParentPath(filename));
}
