#include "sharedfolders.h"
#include <QSettings>
#include <QDebug>
#include "globals.h"

using namespace qshare;

SharedFolders *SharedFolders::pInstance = NULL;
const QString SharedFolders::separator('/');

SharedFolders *SharedFolders::instance()
{
	if (pInstance)
		return pInstance;
	return new SharedFolders;
}

SharedFolders::Folder::Folder(const QString &path, bool writable) : writable(writable), path(path)
{
}

bool SharedFolders::Folder::isWritable() const
{
	return writable;
}

const QString &SharedFolders::Folder::getPath() const
{
	return path;
}

void SharedFolders::Folder::setWritable(bool writable)
{
	this->writable = writable;
}

void SharedFolders::Folder::setPath(const QString &path)
{
	this->path = path;
}

SharedFolders::SharedFolders() : QMutex(QMutex::Recursive)
{
	pInstance = this;
	loadFolders();
}

SharedFolders::~SharedFolders()
{
	pInstance = NULL;
}

void SharedFolders::loadFolders()
{
	QMutexLocker mLock(this);
	int n = settings->beginReadArray("folders");
	folders.clear();
	for(int i = 0 ; i < n ; ++i)
	{
		settings->setArrayIndex(i);
		Folder folder(settings->value("path").toString(),
					  settings->value("writable").toBool());
		folders[settings->value("name").toString()] = folder;
	}
	settings->endArray();
}

void SharedFolders::saveFolders()
{
	QMutexLocker mLock(this);
	settings->beginWriteArray("folders");
	int e(0);
	for(QMap<QString, Folder>::iterator i = folders.begin() ; i != folders.end() ; ++i)
	{
		settings->setArrayIndex(e++);
		settings->setValue("path", i.value().getPath());
		settings->setValue("writable", i.value().isWritable());
		settings->setValue("name", i.key());
	}
	settings->endArray();
}

void SharedFolders::addFolder(const QString &name, const Folder &folder)
{
	QMutexLocker mLock(this);
	folders[name] = folder;
	saveFolders();
}

void SharedFolders::removeFolder(const QString &name)
{
	QMutexLocker mLock(this);
	folders.remove(name);
	saveFolders();
}

QMap<QString, SharedFolders::Folder> SharedFolders::getFolderList()
{
	QMutexLocker mLock(this);
	return folders;
}

QString SharedFolders::getRealPath(const QString &path)
{
	QStringList keys = path.split(separator, QString::SkipEmptyParts);
	if (keys.isEmpty())
		return separator;

	QMutexLocker mLock(this);
	if (folders.contains(keys[0]))
	{
		keys[0] = folders[keys[0]].getPath();
		return keys.join(separator);
	}
	return separator;
}

bool SharedFolders::isWritable(const QString &path)
{
	QStringList keys = path.split(separator, QString::SkipEmptyParts);
	if (keys.isEmpty())
		return false;

	QMutexLocker mLock(this);
	if (folders.contains(keys[0]))
	{
		return folders[keys[0]].isWritable();
	}
	return false;
}

int SharedFolders::nbFolders()
{
	return folders.size();
}

QString SharedFolders::getParentPath(const QString &path)
{
	QStringList keys = path.split(separator, QString::SkipEmptyParts);
	if (keys.size() <= 1)
		return separator;
	keys.removeLast();
	return separator + keys.join(separator);
}

QString SharedFolders::getFileName(const QString &path)
{
	QStringList keys = path.split(separator, QString::SkipEmptyParts);
	if (keys.isEmpty())
		return QString();
	return keys.last();
}
