#include "stats.h"
#include "ui_stats.h"
#include "ftpserver.h"
#include <QTimer>
#include <QGraphicsScene>
#include <QRgb>
#include <QDebug>
#include "utils.h"
#include "config.h"

Stats *Stats::pInstance = NULL;

Stats *Stats::instance()
{
	if (pInstance)
		return pInstance;
	return new Stats;
}

void Stats::deleteInstance()
{
	if (pInstance)
		pInstance->deleteLater();
}

Stats::Stats(QWidget *parent) :
    QDialog(parent),
    m_ui(new Ui::Stats)
{
	pInstance = this;

    m_ui->setupUi(this);

	m_ui->chart->setScene(new QGraphicsScene);

	prevSent = 0;
	prevRecv = 0;

	QTimer *timer = new QTimer;
	connect(timer,SIGNAL(timeout()),this,SLOT(updateChart()));
	connect(timer,SIGNAL(timeout()),m_ui->stats,SLOT(repaint()));
	timer->setInterval(250);
	timer->setSingleShot(false);
	timer->start();
	clock.start();

	setWindowFlags(Qt::Dialog | Qt::WindowStaysOnTopHint);
    setWindowTitle(windowTitle() + " (" + Config::instance()->getHostname() + ')');
}

Stats::~Stats()
{
    delete m_ui;

	pInstance = NULL;
}

void Stats::changeEvent(QEvent *e)
{
    QDialog::changeEvent(e);
    switch (e->type()) {
    case QEvent::LanguageChange:
        m_ui->retranslateUi(this);
        setWindowTitle(windowTitle() + " (" + Config::instance()->getHostname() + ')');
        break;
    default:
        break;
    }
}

void Stats::showEvent(QShowEvent *)
{
	updateStats();
}

void Stats::updateStats()
{
	if (!isVisible())
		return;

	if (!FtpServer::serverRunning())
	{
		m_ui->users->setRowCount(0);
		return;
	}

	setUpdatesEnabled(false);
	QList<FtpServer::ConnectionInfo> lInfo = FtpServer::getClientsInfo();
	m_ui->users->setRowCount(lInfo.size());
	int i(0);
	foreach(FtpServer::ConnectionInfo info, lInfo)
	{
		m_ui->users->setItem(i, 0, new QTableWidgetItem(info.user));
		m_ui->users->setItem(i, 1, new QTableWidgetItem(info.address));
		m_ui->users->setItem(i, 2, new QTableWidgetItem(info.activity));
		++i;
	}

	QStringList sInfo = FtpServer::getServerInfo();
	for(int i = 0 ; i < sInfo.size() ; ++i)			// Make sure there are items there
		if (!m_ui->stats->item(i,0))
			m_ui->stats->setItem(i, 0, new QTableWidgetItem());

	quint64 size = sInfo[0].toULongLong();			// Sent bytes
	QString bytesStr;
	if (!(size >> 10))
		bytesStr = QString::number(size) + ' ' + tr("Bytes");
	else if (!(size >> 20))
		bytesStr = QString::number(size / 1024.0, 'f', 3) + ' ' + tr("KB");
	else if (!(size >> 30))
		bytesStr = QString::number((size >> 10) / 1024.0, 'f', 3) + ' ' + tr("MB");
	else if (!(size >> 40))
		bytesStr = QString::number((size >> 20) / 1024.0, 'f', 3) + ' ' + tr("GB");
	else
		bytesStr = QString::number((size >> 30) / 1024.0, 'f', 3) + ' ' + tr("TB");
	m_ui->stats->item(0,0)->setText(bytesStr);

	size = sInfo[1].toULongLong();				// Received bytes
	bytesStr.clear();
	if (!(size >> 10))
		bytesStr = QString::number(size) + ' ' + tr("Bytes");
	else if (!(size >> 20))
		bytesStr = QString::number(size / 1024.0, 'f', 3) + ' ' + tr("KB");
	else if (!(size >> 30))
		bytesStr = QString::number((size >> 10) / 1024.0, 'f', 3) + ' ' + tr("MB");
	else if (!(size >> 40))
		bytesStr = QString::number((size >> 20) / 1024.0, 'f', 3) + ' ' + tr("GB");
	else
		bytesStr = QString::number((size >> 30) / 1024.0, 'f', 3) + ' ' + tr("TB");
	m_ui->stats->item(1,0)->setText(bytesStr);
	for(int i = 2 ; i < sInfo.size() ; ++i)
		m_ui->stats->item(i, 0)->setText(sInfo[i]);

	setUpdatesEnabled(true);
}

void Stats::updateChart()
{
	QStringList sInfo = FtpServer::getServerInfo();

	quint64 elapsed = clock.restart();
	quint64 timeCoef = (quint64(250) << 32) / elapsed;

	if (timeline.isEmpty())
	{
		timeline.push_back(0.0);
		connections.push_back(0);
		bytesOut.push_back(0);
		bytesIn.push_back(0);
	}

	timeline.push_back(timeline.back() + elapsed / 250.0);
	connections.push_back(sInfo[2].toInt());
	const quint64 vout = (sInfo[0].toLongLong() - prevSent) * timeCoef >> 32;
	const quint64 vin = (sInfo[1].toLongLong() - prevRecv) * timeCoef >> 32;
	bytesOut.push_back((bytesOut.back() + vout) >> 1);
	bytesIn.push_back((bytesIn.back() + vin) >> 1);

	prevSent = sInfo[0].toULongLong();
	prevRecv = sInfo[1].toULongLong();
	const int maxlen = m_ui->chart->width() - 8;
	while (timeline.back() - timeline.front() < maxlen)
	{
		timeline.push_front(timeline.front() - 1.0);
		bytesIn.push_front(0);
		bytesOut.push_front(0);
		connections.push_front(0);
	}

	while (timeline.back() - timeline.front() > maxlen)
	{
		timeline.pop_front();
		bytesIn.pop_front();
		bytesOut.pop_front();
		connections.pop_front();
	}

	if (!isVisible())
		return;

	setUpdatesEnabled(false);
	m_ui->chart->scene()->clear();
	const double start = timeline.front();
	qint64 m = qMax<qint64>(1, max(bytesOut));
	double coef = -0.667 * m_ui->chart->height() / double(m);
	QPolygonF polyOut;
	for(int i = 0 ; i < bytesOut.size() ; ++i)
		polyOut << QPointF(timeline[i] - start, bytesOut[i] * coef);
	polyOut << QPointF(timeline.back() - start, 0.0) << QPointF(0.0, 0.0);

	m = qMax<qint64>(1, max(bytesIn));
	coef = -0.667 * m_ui->chart->height() / double(m);
	QPolygonF polyIn;
	for(int i = 0 ; i < bytesOut.size() ; ++i)
		polyIn << QPointF(timeline[i] - start, bytesIn[i] * coef);
	polyIn << QPointF(timeline.back() - start, 0.0) << QPointF(0.0, 0.0);

	m = qMax<qint64>(1, max(connections));
	coef = -0.667 * m_ui->chart->height() / double(m);
	QPolygonF polyCon;
	for(int i = 0 ; i < bytesOut.size() ; ++i)
		polyCon << QPointF(timeline[i] - start, connections[i] * coef);
	polyCon << QPointF(timeline.back() - start, 0.0) << QPointF(0.0, 0.0);

	m_ui->chart->scene()->addPolygon(polyCon, QPen(QColor(0,255,0)));
	m_ui->chart->scene()->addPolygon(polyOut, QPen(QColor(255,0,0)), QBrush(QColor(255,0,0,128)));
	m_ui->chart->scene()->addPolygon(polyIn, QPen(QColor(0,0,255)), QBrush(QColor(0,0,255,128)));

	setUpdatesEnabled(true);
}
